import multiprocessing
from openQCM.core.ringBuffer import RingBuffer
from openQCM.core.constants import Constants
from openQCM.common.fileStorage import FileStorage
from openQCM.common.logger import Logger as Log
from openQCM.common.switcher import Overtone_Switcher_5MHz, Overtone_Switcher_10MHz
from time import time
import serial
from serial.tools import list_ports
import numpy as np
from numpy import loadtxt
from scipy.interpolate import UnivariateSpline

from progressbar import Bar, Percentage, ProgressBar, RotatingMarker,Timer

TAG = ""#"[Serial]"

###############################################################################
# Process for the serial package and the communication with the serial port
# Processes incoming data and calculates outgoing data by the algorithms
###############################################################################
class SerialProcess(multiprocessing.Process):
    
    
    ###########################################################################
    # BASELINE CORRECTION
    ###########################################################################
    def baseline_correction(self,x,y,poly_order):
        
        # Estimate Baseline with Least Squares Polynomial Fit (LSP)
        coeffs = np.polyfit(x,y,poly_order) 
        # Evaluate a polynomial at specific values
        poly_fitted = np.polyval(coeffs,x) 
        return poly_fitted,coeffs       
      
    ###########################################################################
    # BASELINE - Evaluates polynomial coefficients - Sweep over all frequencies
    ###########################################################################
    def baseline_coeffs(self):
        
        # initializations
        self.polyfitted_all = None
        self.coeffs_all = None
        self.polyfitted_all_phase = None
        self.coeffs_all_phase = None
        
        # loads Calibration (baseline correction) from file
        (self.freq_all,self.mag_all,self.phase_all) = self.load_calibration_file()
        
        # Baseline correction: input signal Amplitude (sweep all frequencies)
        (self.polyfitted_all,self.coeffs_all)=self.baseline_correction(self.freq_all,self.mag_all,8)
        self.mag_beseline_corrected_all= self.mag_all-self.polyfitted_all
        
        # Baseline correction: input signal Phase (sweep all frequencies)
        (self.polyfitted_all_phase,self.coeffs_all_phase)=self.baseline_correction(self.freq_all,self.phase_all,8)
        self.phase_beseline_corrected_all= self.phase_all-self.polyfitted_all_phase 
        return self.coeffs_all
    
    
    ###########################################################################
    # Savitzky-Golay (Smoothing/Denoising Filter)
    ###########################################################################
    def savitzky_golay(self,y, window_size, order, deriv=0, rate=1):
        
        """Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
        The Savitzky-Golay filter removes high frequency noise from data.
        It has the advantage of preserving the original shape and
        features of the signal better than other types of filtering
        approaches, such as moving averages techniques.
        Parameters
        ----------
        y : array_like, shape (N,) the values of the time history of the signal.
        window_size : int the length of the window. Must be an odd integer number.
        order : int the order of the polynomial used in the filtering.
                Must be less then `window_size` - 1.
        deriv: int the order of the derivative to compute (default = 0 means only smoothing)
        Returns
        -------
        ys : ndarray, shape (N) the smoothed signal (or it's n-th derivative).
        Notes
        -----
        The Savitzky-Golay is a type of low-pass filter, particularly
        suited for smoothing noisy data. The main idea behind this
        approach is to make for each point a least-square fit with a
        polynomial of high order over a odd-sized window centered at
        the point.
        Examples
        --------
        t = np.linspace(-4, 4, 500)
        y = np.exp( -t**2 ) + np.random.normal(0, 0.05, t.shape)
        ysg = savitzky_golay(y, window_size=31, order=4)
        import matplotlib.pyplot as plt
        plt.plot(t, y, label='Noisy signal')
        plt.plot(t, np.exp(-t**2), 'k', lw=1.5, label='Original signal')
        plt.plot(t, ysg, 'r', label='Filtered signal')
        plt.legend()
        plt.show()
        References
        ----------
        .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of
           Data by Simplified Least Squares Procedures. Analytical
           Chemistry, 1964, 36 (8), pp 1627-1639.
        .. [2] Numerical Recipes 3rd Edition: The Art of Scientific Computing
           W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
           Cambridge University Press ISBN-13: 9780521880688
        """
        import numpy as np
        from math import factorial
        try:
            window_size = np.abs(np.int(window_size))
            order = np.abs(np.int(order)) 
        except ValueError as msg:
            raise ValueError("WARNING: window size and order have to be of type int!")
        if window_size % 2 != 1 or window_size < 1:
            raise TypeError("WARNING: window size must be a positive odd number!")
        if window_size < order + 2:
            raise TypeError("WARNING: window size is too small for the polynomials order!")
        order_range = range(order+1)
        half_window = (window_size -1) // 2
        # precompute coefficients
        b = np.mat([[k**i for i in order_range] for k in range(-half_window, half_window+1)])
        m = np.linalg.pinv(b).A[deriv] * rate**deriv * factorial(deriv)
        # pad the signal at the extremes with values taken from the signal itself
        firstvals = y[0] - np.abs( y[1:half_window+1][::-1] - y[0] )
        lastvals = y[-1] + np.abs(y[-half_window-1:-1][::-1] - y[-1])
        y = np.concatenate((firstvals, y, lastvals))
        return np.convolve( m[::-1], y, mode='valid')
     
        
    ###########################################################################
    # Resonance Frequency, Resonance Peak, Bandwidth and Q-factor/Dissipation 
    ###########################################################################
    def parameters_finder(self,freq,signal,percent):
        f_max = np.max(signal)          # Find maximum
        i_max= np.argmax(signal,axis=0) # Find index of maximum
        # setup the index for finding the leading edge
        index_m = i_max
        # loop until the index at FWHM/others is found
        while signal[index_m] > percent*f_max:
            if index_m < 1:
               #print(TAG, 'WARNING: Left value not found')
               self._err1 = 1
               break
            index_m = index_m-1     
        #linearly interpolate between the previous values to find the value of freq at the leading edge
        m = (signal[index_m+1] - signal[index_m])/(freq[index_m+1] - freq[index_m])
        c = signal[index_m] - freq[index_m]*m
        i_leading = (percent*f_max - c)/m
        # setup index for finding the trailing edge
        index_M = i_max
        # loop until the index at FWHM/others is found
        while signal[index_M] > percent*f_max:
            if index_M >= len(signal)-1:
                #print(TAG, 'WARNING: Right value not found')
                self._err2 = 1
                break
            index_M = index_M+1;
        # linearly interpolate between the previous values to find the value of freq at the trailing edge
        m = (signal[index_M-1] - signal[index_M])/(freq[index_M-1] - freq[index_M])
        c = signal[index_M] - freq[index_M]*m
        i_trailing = (percent*f_max - c)/m
        #compute the FWHM/others
        bandwidth = abs(i_trailing - i_leading)
        Qfac=freq[i_max]/bandwidth
        return i_max, f_max, bandwidth, index_m, index_M, Qfac  
    
    
    ###########################################################################
    # Processes incoming data and calculates outcoming data
    ###########################################################################    
    def elaborate(self, k, coeffs_all, readFREQ, samples, Xm, Xp, temperature, SG_window_size, Spline_points, Spline_factor, timestamp):
        
        ###################
        def waveletSmooth(x, wavelet="db4", level=1, title=None):
            import pywt
            from statsmodels.robust import mad
            # calculate the wavelet coefficients
            coeff = pywt.wavedec( x, wavelet, mode="per")
            # calculate a threshold
            sigma = mad(coeff[-level])
            # changing this threshold also changes the behavior
            uthresh = sigma * np.sqrt( 2*np.log(len(x)))
            coeff[1:] = (pywt.threshold(i, value=uthresh, mode="soft") for i in coeff[1:])
            # reconstruct the signal using the thresholded coefficients
            y = pywt.waverec( coeff, wavelet, mode="per")
            return y
        ###################
        # Number of spline points
        points = Spline_points
        # sweep counter
        self._k= k
        # evaluated polynomial coefficients
        self._coeffs_all = coeffs_all
        # frequency range, samples number
        self._readFREQ = readFREQ
        self._samples = samples
        # support vectors
        self._Xm = Xm
        self._Xp = Xp
        self._filtered_mag = np.zeros(samples)
        # save current data 
        mag   = self._Xm
        phase = self._Xp ############################
        # Initializations of support vectors for later storage
        self._Xm = np.linspace(0,0,self._samples)
        self._Xp = np.linspace(0,0,self._samples)
        
        # Evaluate a polynomial at specific values based on the coefficients and frequency range
        self._polyfitted = np.polyval(self._coeffs_all,self._readFREQ)
        
        # BASELINE CORRECTION ROI (raw data)
        mag_beseline_corrected = mag-self._polyfitted
        
        # FILTERING - Savitzky-Golay
        filtered_mag = self.savitzky_golay(mag_beseline_corrected, window_size = SG_window_size, order = Constants.SG_order)
        
        # peak, index e frequency of max detection baseline corrected (filtering optional)
        #self._vector_max_baseline_corrected.append(max(mag_beseline_corrected))   #Z axis (max)
        #self._index_max_baseline_corrected.append(np.argmax(mag_beseline_corrected, axis=0)) # X axis (max position)
        #h=self._index_max_baseline_corrected.append(np.argmax(mag_beseline_corrected, axis=0))
        #self._freq_max_baseline_corrected.append(readFREQ[int(h)])
        
        # FITTING/INTERPOLATING - SPLINE
        xrange = range(len(filtered_mag))
        freq_range = np.linspace(self._readFREQ[0], self._readFREQ[-1], points)
        s = UnivariateSpline(xrange, filtered_mag, s= Spline_factor)
        xs = np.linspace(0, len(filtered_mag)-1, points)
        mag_result_fit = s(xs)
        
        # PARAMETERS FINDER
        (index_peak_fit, max_peak_fit, bandwidth_fit,index_f1_fit,index_f2_fit, Qfac_fit)= self.parameters_finder(freq_range, mag_result_fit, percent=0.707)
        
        # BANDWIDTH 70.7% of MAX
        #self._bw3.append(bandwidth_fit)
        # Q FACTOR/DISSIPATION
        #self._q_factor.append (1/Qfac_fit)
        # Index of MAX fitted signals ()
        #self._index_max_fit[k]= int(index_peak_fit)
        # Frequency of MAX fitted signals ()
        #self._freq_max_fit.append(freq_range[int(index_peak_fit)])
        #self._temperature.append(temperature)
        #######################################################
        
        self._frequency_buffer.append(freq_range[int(index_peak_fit)])
        self._dissipation_buffer.append(1/Qfac_fit)
        self._temperature_buffer.append(temperature)
        
        if self._k>=self._environment:
           #FREQUENCY
           vec_app1 = self.savitzky_golay(self._frequency_buffer.get_all(), window_size = Constants.SG_window_environment, order = Constants.SG_order_environment)
           freq_range_mean = np.average(vec_app1)
           #DISSIPATION     
           vec_app1d = self.savitzky_golay(self._dissipation_buffer.get_all(), window_size = Constants.SG_window_environment, order = Constants.SG_order_environment)
           diss_mean = np.average(vec_app1d)
           #TEMPERATURE
           vec_app1t = self.savitzky_golay(self._temperature_buffer.get_all(), window_size = Constants.SG_window_environment, order = Constants.SG_order_environment)
           temperature_mean = np.average(vec_app1t)
           
        #else:
           #freq_range_mean = freq_range[int(index_peak_fit)]
           #temperature_mean = temperature
           #diss_mean = 1/Qfac_fit
           #freq_range_mean = freq_range[int(index_peak_fit)]
           #diss_mean = 1/Qfac_fit
           #temperature_mean = temperature
        # vector of MAX fitted signals ()
        #self._vector_max_fit.append(max_peak_fit)
        
        #######################################################
        ##############
        import datetime
        epoch= datetime.datetime(1970, 1, 1, 0, 0) #offset-naive datetime
        ts_mult=1e6
        w = (int((datetime.datetime.now() - epoch).total_seconds()*ts_mult)) #datetime.datetime.utcnow()
        ##############
        ## ADDS new serial data to internal queue
        self._parser1.add1(filtered_mag) ##############
        self._parser2.add2(phase)        ##############
        # Adds new calculated data (resonance frequency and dissipation) to internal queues
        #self._parser3.add3([time()-timestamp,freq_range[int(index_peak_fit)]])
        self._parser3.add3([w,freq_range_mean]) #time()-timestamp - time in seconds
        #self._parser4.add4([time()-timestamp,1/Qfac_fit])
        self._parser4.add4([w,diss_mean]) #time()-timestamp - time in seconds
        #self._parser5.add5([time()-timestamp,temperature])
        self._parser5.add5([w,temperature_mean]) #time()-timestamp - time in seconds
        '''
        ##############################
        # DATA STORING in CSV/TXT FILE
        ##############################
        import csv
        from time import strftime, localtime
        # STORING DATA (CSV) in main folder: Time,resonance Frequency,Dissipation
        filename = 'data_q.csv'
        with open(filename,'a', newline='') as tempFile:
         tempFileWriter = csv.writer(tempFile)
         tempFileWriter.writerow([strftime(Constants.csv_default_filename, localtime()),freq_range[int(index_peak_fit)],1/Qfac_fit])
        tempFile.close() 
        '''
        '''
        import csv
        # STORING DATA (CSV) in main folder: Time,resonance Frequency,Dissipation
        filename = 'bw3.csv'
        with open(filename,'a', newline='') as tempFile:
         tempFileWriter = csv.writer(tempFile)
         tempFileWriter.writerow(self._bw3)
        tempFile.close()
        
        filename1 = 'vector_max.csv'
        with open(filename1,'a', newline='') as tempFile1:
         tempFileWriter = csv.writer(tempFile1)
         tempFileWriter.writerow(self._vector_max_fit)
        tempFile.close()
        '''
        # STORING DATA (CSV/TXT) in 'data' folder: frequency, amplitude
        #np.savetxt(r"logged_data\sweep_%d_mag_raw.txt"%(k,), np.column_stack([readFREQ,mag]))
        #np.savetxt(r"logged_data\data\sweep_%d_mag_baseline_corrected.txt"%(k,), np.column_stack([readFREQ,mag_beseline_corrected]))
        #np.savetxt(r"data\sweep_%d_mag_baseline_corrected.csv"%(k,), np.column_stack([readFREQ,globals()["Af_baseline_" + str(k)]]), delimiter=',')
        #np.savetxt(r"logged_data\data\sweep_%d_mag_filtered.txt"%(k,), np.column_stack([readFREQ,filtered_mag]))
        #np.savetxt(r"logged_data\data\sweep_%d_mag_fitted.txt"%(k,), np.column_stack([freq_range,mag_result_fit]))
        ##############
        

    ###########################################################################
    # Initializing values for process
    ###########################################################################
    def __init__(self, parser_process):
        """
        :param parser_process: Reference to a ParserProcess instance.
        :type parser_process: ParserProcess.
        """
        multiprocessing.Process.__init__(self)
        self._exit = multiprocessing.Event()
        
        # Instantiate a ParserProcess class for each communication channels
        self._parser1 = parser_process
        self._parser2 = parser_process
        self._parser3 = parser_process
        self._parser4 = parser_process
        self._parser5 = parser_process
        self._parser6 = parser_process
        self._serial = serial.Serial()
        
    ###########################################################################
    # Opens a specified serial port
    ###########################################################################    
    def open(self, port, 
                   speed = Constants.serial_default_overtone, 
                   timeout = Constants.serial_timeout_ms, 
                   writeTimeout = Constants.serial_writetimeout_ms):
        """
        :param port: Serial port name :type port: str.
        :param speed: Overtone selected for the analysis :type speed: str.
        :param timeout: Sets current read timeout :type timeout: float (seconds).
        :param writetTimeout: Sets current write timeout :type writeTimeout: float (seconds).
        :return: True if the port is available :rtype: bool.
        """
    
        self._serial.port = port
        self._serial.baudrate = Constants.serial_default_speed #115200
        self._serial.stopbits = serial.STOPBITS_ONE
        self._serial.bytesize = serial.EIGHTBITS
        self._serial.timeout = timeout
        self._serial.writetimeout = writeTimeout
        
        #self._overtone = float(speed)
        # Loads frequencies from file
        peaks_mag = self.load_frequencies_file()
        
        # handles the exceptions
        try:
           self._overtone = float(speed)
        except:   
           print(TAG, "Warning: wrong frequency selection, set default to {} Hz Fundamental".format(peaks_mag[0]))
           self._overtone = peaks_mag[0]
    
        self._overtone_int = None
        for i in range(len(peaks_mag)):
            if self._overtone == peaks_mag[i]:
               self._overtone_int = i
        # Checks for correct frequency selection
        if self._overtone_int == None:
           print(TAG, "Warning: wrong frequency selection, set default to {} Hz Fundamental".format(peaks_mag[0])) 
           self._overtone_int = 0

        return self._is_port_available(self._serial.port)
    
    ###########################################################################
    # Reads the serial port,processes and adds all the data to internal queues
    ###########################################################################
    def run(self):
        """
        The expected format is a buffer (sweep) and a new buffer as a new sweep. 
        The method parses data, converts each value to float and adds to a queue. 
        If incoming data can't be converted to float,the data will be discarded.
        """  
        # initializations
        #self._vector_max_baseline_corrected = []
        #self._index_max_baseline_corrected = []
        #self._freq_max_baseline_corrected = []
        #self._vector_max_fit = []
        #self._index_max_fit = []
        #self._bw3 = []
        #self._q_factor = []
        #self._freq_max_fit = []
        #self._temperature = []
        self._flag_error = 0
        self._flag_error_usb = 0
        self._err1 = 0
        self._err2 = 0
              
        # CALLS baseline_coeffs method
        coeffs_all = self.baseline_coeffs()
        
        # Checks if the serial port is currently connected
        if self._is_port_available(self._serial.port):

            samples = Constants.argument_default_samples 
            # Calls get_frequencies method:
            # ACQUIRES overtone, sets start and stop frequencies, the step and range frequency according to the number of samples
            (overone_name,overtone_value,fStep,readFREQ,SG_window_size,Spline_points,Spline_factor) = self.get_frequencies(samples)
            
            # Gets the state of the serial port
            if not self._serial.isOpen(): 
                # OPENS the serial port
                self._serial.open() 
                # Initializes the sweep counter
                k=0 
                print(TAG,'Capturing raw data...')
                print(TAG,'Wait, processing early data...')
                # creates a timestamp
                timestamp = time()
                
                self._environment = Constants.environment
                self._frequency_buffer   = RingBuffer(self._environment)
                self._dissipation_buffer = RingBuffer(self._environment)
                self._temperature_buffer = RingBuffer(self._environment)
                # Initializes the progress bar  
                bar = ProgressBar(widgets=[TAG,' ', Bar(marker='>'),' ',Percentage(),' ', Timer()], maxval=self._environment).start() #
                #### SWEEPS LOOP ####
                while not self._exit.is_set():
                    # data reset for new sweep 
                    data_mag = np.linspace(0,0,samples)   
                    data_ph  = np.linspace(0,0,samples)
                    
                    try:
                        # amplitude/phase convert bit to dB/Deg parameters
                        vmax = 3.3
                        bitmax = 8192 
                        ADCtoVolt = vmax / bitmax
                        VCP = 0.9
                        
                        # WRITES encoded command to the serial port
                        cmd = str(self._startFreq) + ';' + str(self._stopFreq) + ';' + str(int(fStep)) + '\n'
                        self._serial.write(cmd.encode())
                        
                        # Initializes buffer and strs record
                        buffer = ''
                        strs = ["" for x in range(samples + 2)]
                        
                        # READS and decodes sweep from the serial port
                        while 1:
                         buffer += self._serial.read(self._serial.inWaiting()).decode(Constants.app_encoding)
                         #if '\n' in buffer:
                         if 's' in buffer:
                              break
                        data_raw = buffer.split('\n')
                        length = len(data_raw)
                        
                        # PERFORMS split with the semicolon delimiter
                        for i in range (length):
                            strs[i] = data_raw[i].split(';')

                        # CONVERTS the sweep samples before adding to queue
                        for i in range (length - 2):
                            data_mag[i] = float(strs[i][0]) * ADCtoVolt / 2
                            data_mag[i] = (data_mag[i]-VCP) / 0.03
                            data_ph[i] = float(strs[i][1]) * ADCtoVolt / 1.5
                            data_ph[i] = (data_ph[i]-VCP) / 0.01
                        
                        # ACQUIRES the temperature value from the buffer 
                        data_temp = float((strs[length - 2][0]))
                            
                    # specify handlers for different exceptions        
                    except ValueError:
                        print(TAG, "WARNING (ValueError): convert raw to float failed", end='\r')
                        #Log.w(TAG, "Warning (ValueError): convert Raw to float failed")
                    except:
                        #if self._flag_error_usb == 1:
                        print(TAG, "WARNING (ValueError): convert raw to float failed", end='\r')
                        self._flag_error_usb += 1
                        #Log.w(TAG, "Warning (ValueError): convert Raw to float failed")
                        
                    ## ADDS new serial data to internal queue
                    #self._parser1.add1(data_mag)
                    #self._parser2.add2(data_ph)
                    
                    # Calls elaborate method to performs results
                    try:
                        self.elaborate(k, coeffs_all, readFREQ, samples,data_mag, data_ph, data_temp, SG_window_size, Spline_points, Spline_factor, timestamp)
                    except ValueError:
                        self._flag_error = 1
                        #if k > self._environment:
                        #   print(TAG, "WARNING (ValueError): miscalculation")
                    except:
                        self._flag_error = 1
                        #if k > self._environment:
                        #   print(TAG, "WARNING (ValueError): miscalculation")
                    self._parser6.add6([self._err1,self._err2,k,self._flag_error_usb])
                    if k<= self._environment:
                       bar.update(k)
                    elif k/50 == k//50:
                      if k==100:
                         print('\n')
                      print(TAG,"sweep #{}               ".format(k), end='\r')
                    # refreshes error variables at each sweep
                    self._err1 = 0
                    self._err2 = 0
                    # Increases sweep counter 
                    k+=1
                if k== self._environment:
                   bar.finish()
                #### END SWEEPS LOOP ####    
                # CLOSES serial port
                self._serial.close()
          
    ###########################################################################
    # Stops acquiring data
    ###########################################################################
    def stop(self):
        #Signals the process to stop acquiring data.
        self._exit.set()
        
        
    ###########################################################################    
    # Automatically selects the serial ports for Teensy (macox/windows)
    ###########################################################################
    @staticmethod
    def get_ports(): 
        from openQCM.common.architecture import Architecture,OSType
        if Architecture.get_os() is OSType.macosx:
            import glob
            return glob.glob("/dev/tty.usbmodem*")
        elif Architecture.get_os() is OSType.linux:
            import glob
            return glob.glob("/dev/ttyACM*")
        else:
            found_ports = []
            port_connected = []
            found = False
            ports_avaiable = list(list_ports.comports())
            for port in ports_avaiable:
                if port[2].startswith("USB VID:PID=16C0:0483"):
                    found = True
                    port_connected.append(port[0])
                #else:
                #    Gets a list of the available serial ports.
                #    found_ports.append(port[0])
            if found:
               found_ports = port_connected 
            return found_ports


    ###########################################################################
    # Gets a list of the Overtones reading from file
    ###########################################################################

    @staticmethod
    def get_speeds():
        #:return: List of the Overtones :rtype: str list.
        # Loads frequencies from  file (path: 'common\')
        data  = loadtxt(Constants.cvs_peakfrequencies_path)
        peaks_mag = data[:,0]
        reversed_peaks_mag = peaks_mag[::-1]
        return [str(v) for v in reversed_peaks_mag]

    ###########################################################################
    # Checks if the serial port is currently connected
    ###########################################################################
    def _is_port_available(self, port):
        """
        :param port: Port name to be verified.
        :return: True if the port is connected to the host :rtype: bool.
        """
        for p in self.get_ports():
            if p == port:
                return True
        return False

    ###########################################################################
    # Sets frequency range for the corresponding overtone 
    ###########################################################################
    def get_frequencies(self, samples):
        """
        :param samples: Number of samples :type samples: int.
        :return: overtone :rtype: float.
        :return: fStep, frequency step  :rtype: float.
        :return: readFREQ, frequency range :rtype: float list.
        """
        # Loads frequencies from file
        peaks_mag = self.load_frequencies_file()
       
        # Checks QCS type 5Mhz or 10MHz
        # Sets start and stop frequencies for the corresponding overtone
        if len(peaks_mag) == 5:
            switch = Overtone_Switcher_5MHz(peak_frequencies = peaks_mag)
            # 0=fundamental, 1=3th overtone and so on
            (overtone_name,overtone_value, self._startFreq,self._stopFreq,SG_window_size,spline_factor) = switch.overtone5MHz_to_freq_range(self._overtone_int)
            print(TAG,"openQCM Device setup: @5MHz")
        elif len(peaks_mag) == 3:
            switch = Overtone_Switcher_10MHz(peak_frequencies = peaks_mag)
            (overtone_name, overtone_value, self._startFreq,self._stopFreq,SG_window_size,spline_factor) = switch.overtone10MHz_to_freq_range(self._overtone_int)
            print(TAG,"openQCM Device setup: @10MHz")
        
        # Sets the frequency step 
        fStep = (self._stopFreq-self._startFreq)/(samples-1)
        
        # Sets spline points for fitting
        spline_points = int((self._stopFreq-self._startFreq))+1
        
        # Sets the frequency range for the corresponding overtone
        readFREQ = np.arange(samples) * (fStep) + self._startFreq 
        return overtone_name, overtone_value, fStep, readFREQ,SG_window_size, spline_points, spline_factor
    
    
    ###########################################################################
    # Loads Fundamental frequency and Overtones from file
    ###########################################################################
    @staticmethod
    def load_frequencies_file():
        data  = loadtxt(Constants.cvs_peakfrequencies_path)
        peaks_mag = data[:,0]
        #peaks_phase = data[:,1] #unused at the moment
        return peaks_mag
        
    
    ###########################################################################
    # Loads Calibration (baseline correction) from file
    ###########################################################################
    def load_calibration_file(self):
        # Loads Fundamental frequency and Overtones from file
        peaks_mag = self.load_frequencies_file()
        
        # Checks QCS type 5Mhz or 10MHz
        if len(peaks_mag) == 5:
           filename = Constants.csv_calibration_path
        elif len(peaks_mag) == 3:
           filename = Constants.csv_calibration_path10 
        data  = loadtxt(filename)
        freq_all  = data[:,0]
        mag_all   = data[:,1]
        phase_all = data[:,2]
        return freq_all, mag_all, phase_all
      
    
# Instantiate the process and run the method 'run' of the class
#a=SerialProcess(multiprocessing.Process)
#a.run()



import multiprocessing
from time import time

import socket

from openQCM.core.constants import Constants
from openQCM.common.logger import Logger as Log


TAG = "Socket"


class SocketProcess(multiprocessing.Process):
    """
    Socket client
    """
    def __init__(self, parser_process):
        """
        Initialises values for process.
        :param parser_process: Reference to a ParserProcess instance.
        :type parser_process: ParserProcess
        """
        multiprocessing.Process.__init__(self)
        self._exit = multiprocessing.Event()
        self._parser = parser_process
        self._socket_client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        Log.i(TAG, "Process Ready")

    def open(self, port='', speed=5555, timeout=0.01):
        """
        Opens a socket connection to specified host and port
        :param port: Host address to connect to.
        :type port: str.
        :param speed: Port number to connect to.
        :type speed: int.
        :param timeout: Sets timeout for socket interactions.
        :type timeout: float.
        :return: True if the connection was open.
        :rtype: bool.
        """
        try:
            #self._socket_client.timeout = timeout
            speed = int(speed)
            self._socket_client.connect((port, speed))
            Log.i(TAG, "Socket open {}:{}".format(port, speed))
            return True
        except socket.timeout:
            Log.w(TAG, "Connection timeout")
        return False

    def run(self):
        """
        Reads the socket until a stop call is made.
        :return:
        """
        Log.i(TAG, "Process starting...")
        timestamp = time()

        while not self._exit.is_set():
            stamp = time() - timestamp
            try:
                data = self._socket_client.recv(Constants.SocketClient.buffer_recv_size).decode()
                if len(data) > 0:
                    self._parser.add([stamp, data])
            except socket.timeout:
                Log.w(TAG, "read timeout")
        Log.i(TAG, "Process finished")

    def stop(self):
        """
        Signals the process to stop acquiring data.
        :return:
        """
        Log.i(TAG, "Process finishing...")
        self._socket_client.close()
        self._exit.set()

    @staticmethod
    def get_default_host():
        """
        Returns a list of local host names, localhost, host name and local ip address, if available.
        :return: str list.
        """
        values = socket.gethostbyaddr(socket.gethostname())
        hostname = values[0]
        hostip = values[2][0]

        if hostip is not None:
            return [Constants.SocketClient.host_default, hostname, hostip]
        else:
            return [Constants.SocketClient.host_default, hostname]

    @staticmethod
    def get_default_port():
        """
        Returns a list of commonly used socket ports.
        :return: str list.
        """
        return [str(v) for v in Constants.SocketClient.port_default]



import multiprocessing
from openQCM.common.logger import Logger as Log


TAG = ""#"[Parser]"

###############################################################################
# Process to parse incoming data and distribute it to worker
###############################################################################
class ParserProcess(multiprocessing.Process):
    
    
    ###########################################################################
    # Initializing values for process
    ###########################################################################
    def __init__(self, data_queue1,
                       data_queue2,
                       data_queue3,
                       data_queue4,
                       data_queue5,
                       data_queue6):
        """
        :param data_queue{i}: References to queue where processed data will be put.
        :type data_queue{i}: multiprocessing Queue.
        """
        multiprocessing.Process.__init__(self)
        self._exit = multiprocessing.Event()

        self._out_queue1 = data_queue1
        self._out_queue2 = data_queue2
        self._out_queue3 = data_queue3
        self._out_queue4 = data_queue4
        self._out_queue5 = data_queue5
        self._out_queue6 = data_queue6

        #print(TAG, 'Process ready')
        #Log.d(TAG, "Process ready")
        
    ###########################################################################
    # Add new raw data and calculated data to the corresponding internal queue
    ###########################################################################
    def add1(self, data):
        """
        Adds new raw data to internal queue1 (serial data: amplitude).
        :param data: Raw data coming from acquisition process.
        :type data: float.
        """
        self._out_queue1.put(data)
        
    def add2(self, data):
        """
        Adds new raw data to internal queue2 (serial data: phase).
        :param data: Raw data coming from acquisition process.
        :type float: float.
        """
        self._out_queue2.put(data)
        
    def add3(self, data):
        """
        Adds new processed data to internal queue3 (Resonance frequency).
        :param data: Calculated data.
        :type data: float.
        """
        self._out_queue3.put(data)
        
    def add4(self, data):
        """
        Adds new processed data to internal queue3 (Q-factor/dissipation).
        :param data: Calculated data.
        :type data: float.
        """
        self._out_queue4.put(data)
        
    def add5(self, data):
        """
        Adds new processed data to internal queue3 (Q-factor/dissipation).
        :param data: Calculated data.
        :type data: float.
        """
        self._out_queue5.put(data)  
    
    def add6(self, data):
        """
        Adds new processed data to internal queue3 (Q-factor/dissipation).
        :param data: Calculated data.
        :type data: float.
        """
        self._out_queue6.put(data) 
        
    def stop(self):
        """
        Signals the process to stop parsing data.
        :return:
        """
        #print(TAG,'Process finishing...')
        #Log.d(TAG, "Process finishing...")
        self._exit.set()    
    '''
    def run(self):
        """
        Process will monitor the internal buffer to parse raw data and distribute to graph and storage, if needed.
        The process will loop again after timeout if more data is available.
        :return:
        """
        print(TAG,'Process starting...')
        Log.d(TAG, "Process starting...")
        #while not self._exit.is_set():
            #while not self._in_queue.empty():
            #      queue = self._in_queue.get(False)#timeout=self._consumer_timeout
            #      self._out_queue.put((queue))
            #      print(queue) #DATI SINCRONI
            #sleep(self._consumer_timeout)
        # last check on the queue to completely remove data.
        #self._consume_queue()
        print(TAG,'Process finished')
        Log.d(TAG, "Process finished")
    '''

    '''
    def run(self):
        """
        Process will monitor the internal buffer to parse raw data and distribute to graph and storage, if needed.
        The process will loop again after timeout if more data is available.
        :return:
        """
        print(TAG,'Process starting...')
        Log.d(TAG, "Process starting...")
        while not self._exit.is_set():
            self._consume_queue()
            sleep(self._consumer_timeout)
        # last check on the queue to completely remove data.
        self._consume_queue()
        print(TAG,'Process finished')
        Log.d(TAG, "Process finished")

    def stop(self):
        """
        Signals the process to stop parsing data.
        :return:
        """
        print(TAG,'Process finishing...')
        Log.d(TAG, "Process finishing...")
        self._exit.set()
        

    def _consume_queue(self):
        """
        Consumer method for the queues/process.
        Used in run method to recall after a stop is requested, to ensure queue is emptied.
        :return:
        """
        while not self._in_queue.empty():
            queue = self._in_queue.get(timeout=self._consumer_timeout)
            self._parse_csv(queue)


    def _parse_csv(self,line):
        """
        Parses incoming data and distributes to external processes.
        :param time: Timestamp.
        :type time: float.
        :param line: Raw data coming from acquisition process.
        :type line: basestring.
        :return:
        """
        self._out_queue.put((line))
        if self._store_reference is not None:
            self._store_reference.add(line)
        '''


from multiprocessing import Queue

from openQCM.core.constants import Constants, SourceType
from openQCM.processors.Parser import ParserProcess
from openQCM.processors.Serial import SerialProcess
from openQCM.processors.SocketClient import SocketProcess
from openQCM.processors.Calibration import CalibrationProcess
from openQCM.common.fileStorage import FileStorage
from openQCM.common.logger import Logger as Log
from openQCM.core.ringBuffer import RingBuffer
import numpy as np
from time import time
#import pywt

TAG = ""#"[Worker]"

###############################################################################
# Service that creates and concentrates all processes to run the application
###############################################################################
class Worker:

    ###########################################################################
    # Creates all processes involved in data acquisition and processing
    ###########################################################################
    def __init__(self,QCS_on = None,
                      port = None,
                      speed = Constants.serial_default_overtone,
                      samples = Constants.argument_default_samples,
                      source = SourceType.serial,
                      export_enabled = False):
        """
        :param port: Port to open on start :type port: str.
        :param speed: Speed for the specified port :type speed: float.
        :param samples: Number of samples :type samples: int.
        :param source: Source type :type source: SourceType.
        :param export_enabled: If true, data will be stored or exported in a file :type export_enabled: bool.
        :param export_path: If specified, defines where the data will be exported :type export_path: str.
        """
        # data queues 
        self._queue1 = Queue()
        self._queue2 = Queue()
        self._queue3 = Queue()
        self._queue4 = Queue()
        self._queue5 = Queue()
        self._queue6 = Queue()
        
        # data buffers
        self._data1_buffer = None 
        self._data2_buffer = None 
        self._d1_buffer = None 
        self._d2_buffer = None
        self._d3_buffer = None
        self._t1_buffer = None 
        self._t2_buffer = None
        self._t3_buffer = None
        self._ser_error1 = 0
        self._ser_error2 = 0
        self._ser_err_usb= 0
        self._control_k = 0
        
        # instances of the processes
        self._acquisition_process = None
        self._parser_process = None
        
        # others
        self._QCS_on = QCS_on # QCS installed on device (unused now)
        self._port = port     # serial port 
        # overtones (str) if 'serial' is called
        # QCS (str) if 'calibration' is called 
        self._speed = speed 
        self._samples = samples
        self._source = source
        self._export = export_enabled
        
        # Supporting variables
        self._d1_store = None # data storing
        self._d2_store = None # data storing
        self._readFREQ = None # frequency range
        self._fStep    = None # sample rate
        self._overtone_name  = None # fundamental/overtones name (str)
        self._overtone_value = None # fundamental/overtones value(float)
        self._count = 0 # sweep counter
        self._flag = True
        self._timestart = 0
        
        
    ###########################################################################
    # Starts all processes, based on configuration given in constructor.
    ###########################################################################
    def start(self):
        
        if self._source == SourceType.serial:
           self._samples = Constants.argument_default_samples
        elif self._source == SourceType.calibration:
           self._samples = Constants.calibration_default_samples
           self._readFREQ = Constants.calibration_readFREQ
        # Setup/reset the internal buffers
        self.reset_buffers(self._samples)
        # Instantiates process
        self._parser_process = ParserProcess(self._queue1,self._queue2,self._queue3,self._queue4,self._queue5,self._queue6)
        # Checks the type of source
        if self._source == SourceType.serial:
            self._acquisition_process = SerialProcess(self._parser_process)
        elif self._source == SourceType.calibration:
            self._acquisition_process = CalibrationProcess(self._parser_process)
        elif self._source == SourceType.SocketClient:
            self._acquisition_process = SocketProcess(self._parser_process)
            
        if self._acquisition_process.open(port=self._port, speed=self._speed):
            if self._source == SourceType.serial:
               (self._overtone_name,self._overtone_value, self._fStep, self._readFREQ, SG_window_size, spline_points, spline_factor) = self._acquisition_process.get_frequencies(self._samples)
               #print(TAG, "Quartz Crystal Sensor installed: {}".format(self._QCS_on))
               print("")
               print(TAG, "DATA MAIN INFORMATION")
               print(TAG, "Selected frequency: {} - {}Hz".format(self._overtone_name,self._overtone_value))
               print(TAG, "Frequency start: {}Hz".format(self._readFREQ[0]))
               print(TAG, "Frequency stop:  {}Hz".format(self._readFREQ[-1]))
               print(TAG, "Frequency range: {}Hz".format(self._readFREQ[-1]-self._readFREQ[0]))
               print(TAG, "Number of samples: {}".format(self._samples-1))
               print(TAG, "Sample rate: {}Hz".format(self._fStep))
               print(TAG, "History buffer size: 180 min\n")
               print(TAG, "MAIN PROCESSING INFORMATION")
               print(TAG, "Method for baseline estimation and correction:")
               print(TAG, "Least Squares Polynomial Fit (LSP)")
               #print(TAG, "Degree of the fitting polynomial: 8")
               print(TAG, "Savitzky-Golay Filtering")
               print(TAG, "Order of the polynomial fit: {}".format(Constants.SG_order))
               print(TAG, "Size of data window (in samples): {}".format(SG_window_size))
               print(TAG, "Oversampling using spline interpolation")
               print(TAG, "Spline points (in samples): {}".format(spline_points-1))
               print(TAG, "Resolution after oversampling: {}Hz".format((self._readFREQ[-1]-self._readFREQ[0])/(spline_points-1)))
               
            elif self._source == SourceType.calibration:
               print("")
               print(TAG, "MAIN CALIBRATION INFORMATION")
               print(TAG, "Calibration frequency start:  {}Hz".format(Constants.calibration_frequency_start))
               print(TAG, "Calibration frequency stop:  {}Hz".format(Constants.calibration_frequency_stop))
               print(TAG, "Frequency range: {}Hz".format(Constants.calibration_frequency_stop-Constants.calibration_frequency_start))
               print(TAG, "Number of samples: {}".format(Constants.calibration_default_samples-1))
               print(TAG, "Sample rate: {}Hz".format(Constants.calibration_fStep))
            print(TAG, 'Training for plot...\n')
            # Starts processes
            self._acquisition_process.start()
            self._parser_process.start()
            return True
        else:
            print(TAG, 'Warning: port is not available')
            Log.i(TAG, "Warning: Port is not available")
            return False


    ###########################################################################
    # Stops all running processes
    ###########################################################################    
    def stop(self):
        
        self._acquisition_process.stop()
        self._parser_process.stop()
        '''
        for process in [self._acquisition_process, self._parser_process]:
            if process is not None and process.is_alive():
                process.stop()
                process.join(Constants.process_join_timeout_ms)
        '''
        print(TAG, 'Running processes stopped...')
        print(TAG, 'Processes finished')
        Log.i(TAG, "Running processes stopped...")
        Log.i(TAG, "Processes finished")
        
        
    ###########################################################################
    # Empties the internal queues, updating data to consumers
    ###########################################################################    
    def consume_queue1(self):
        # queue1 for serial data: amplitude
        while not self._queue1.empty():
            self._queue_data1(self._queue1.get(False))
    
    def consume_queue2(self):
        # queue2 for serial data: phase
        while not self._queue2.empty():
            self._queue_data2(self._queue2.get(False))

    def consume_queue3(self):
        # queue3 for elaborated data: resonance frequency
        while not self._queue3.empty():
            self._queue_data3(self._queue3.get(False))
            
    def consume_queue4(self):
        # queue3 for elaborated data: Q-factor/Dissipation
        while not self._queue4.empty():
            self._queue_data4(self._queue4.get(False))    
           
    def consume_queue5(self):
        # queue3 for elaborated data: Temperature
        while not self._queue5.empty():
            self._queue_data5(self._queue5.get(False)) 
    
    def consume_queue6(self):
        # queue3 for elaborated data: errors
        while not self._queue6.empty():
            self._queue_data6(self._queue6.get(False))
            
    ###########################################################################
    # Adds data to internal buffers.
    ###########################################################################    
    def _queue_data1(self,data):
        #:param data: values to add for serial data: amplitude :type data: float.
        self._data1_buffer = data
    
    #####    
    def _queue_data2(self,data):
        #:param data: values to add for serial data phase :type data: float.
        self._data2_buffer = data
        # Additional function: exports calibration data in a file if export box is checked.
        '''
        self.store_data_calibration()
        '''
    #####
    def _queue_data3(self,data):
        #:param data: values to add for Resonance frequency :type data: float.
        self._t1_store = data[0] # time (unused)
        self._d1_store = data[1] # data
        self._t1_buffer.append(data[0])
        self._d1_buffer.append(data[1])
        
    #####
    def _queue_data4(self,data):
        # Additional function: exports processed data in a file if export box is checked.
        #:param data: values to add for Q-factor/dissipation :type data: float.
        self._t2_store = data[0] # time (unused)
        self._d2_store = data[1] # data
        self._t2_buffer.append(data[0])
        self._d2_buffer.append(data[1])
    
    #####
    def _queue_data5(self,data):
        # Additional function: exports processed data in a file if export box is checked.
        #:param data: values to add for temperature :type data: float.
        self._t3_store = data[0] # time (unused)
        self._d3_store = data[1] # data 
        self._t3_buffer.append(data[0])
        self._d3_buffer.append(data[1])
        # for storing relative time 
        if  self._flag and ~np.isnan(self._d3_store):
            self._timestart=time()
            self._flag = False
        # Data Storage in csv and/or txt file 
        self.store_data()
    
        #####
    def _queue_data6(self,data):
        #:param data: values to add for serial error :type data: float.
        self._ser_error1 = data[0]
        self._ser_error2 = data[1]
        self._control_k = data[2]
        self._ser_err_usb = data[3]
        
    ###########################################################################
    # Gets data buffers for plot (Amplitude,Phase,Frequency and Dissipation) 
    ###########################################################################        
    def get_value1_buffer(self):
        #:return: float list.
        return self._data1_buffer
    
    #####
    def get_value2_buffer(self):
        #:return: float list.
        return self._data2_buffer

    #####
    def get_d1_buffer(self):
        #:return: float list.
        return self._d1_buffer.get_all()
        
    ##### Gets time buffers
    def get_t1_buffer(self):
        #:return: float list.
        return self._t1_buffer.get_all()
    
    #####
    def get_d2_buffer(self):
        #:return: float list.
        return self._d2_buffer.get_all()
    
    ##### Gets time buffers
    def get_t2_buffer(self):
        #:return: float list.
        return self._t2_buffer.get_all()
    
    #####
    def get_d3_buffer(self):
        #:return: float list.
        return self._d3_buffer.get_all()
    
    ##### Gets time buffers
    def get_t3_buffer(self):
        #:return: float list.
        return self._t3_buffer.get_all()
    
    ##### Gets serial error
    def get_ser_error(self):
        #:return: float list.
        return self._ser_error1,self._ser_error2, self._control_k, self._ser_err_usb
    

    ###########################################################################
    # Exports data in csv and/or txt file if export box is checked
    ###########################################################################
    def store_data(self): 
        # Checks the type of source
        if self._source == SourceType.serial:
          # Checks the state of the export box
          #if self._export:
          # Storing calculated data with the format: timestamp,resonance frequency,dissipation
          filenameCSV = "{}_{}".format(Constants.csv_filename, self._overtone_name)
          FileStorage.CSVsave(filenameCSV, Constants.csv_export_path,time()-self._timestart, self._d3_store, self._d1_store, self._d2_store)

          if self._export:   
              # Storing acquired sweeps
              filename = "{}_{}_{}".format(Constants.csv_sweeps_filename, self._overtone_name,self._count)
              #filename = "{}_{}".format(Constants.csv_sweeps_filename,self._count)
              path = "{}_{}".format(Constants.csv_sweeps_export_path, self._overtone_name) 
              #FileStorage.CSV_sweeps_save(filename, path, self._readFREQ, self._data1_buffer, self._data2_buffer)
              FileStorage.TXT_sweeps_save(filename, path, self._readFREQ, self._data1_buffer, self._data2_buffer)
          self._count+=1
    
    ###########################################################################
    # Checks if processes are running
    ###########################################################################
    def is_running(self):  
        #:return: True if a process is running :rtype: bool.
        return self._acquisition_process is not None and self._acquisition_process.is_alive()


    ###########################################################################
    # Gets the available ports for specified source
    ###########################################################################
    @staticmethod
    def get_source_ports(source):
        
        """
        :param source: Source to get available ports :type source: SourceType.
        :return: List of available ports :rtype: str list.
        """
        if source == SourceType.serial:
            print(TAG,'Port connected:',SerialProcess.get_ports())
            return SerialProcess.get_ports()
        elif source == SourceType.calibration:
            print(TAG,'Port connected:',CalibrationProcess.get_ports())
            return CalibrationProcess.get_ports()
        elif source == SourceType.SocketClient:
            return SocketProcess.get_default_host()
        else:
            print(TAG,'Warning: unknown source selected')
            Log.w(TAG,"Unknown source selected")
            return None
        
        
    ###########################################################################
    # Gets the available speeds for specified source
    ###########################################################################
    @staticmethod
    def get_source_speeds(source):
        
        """
        :param source: Source to get available speeds :type source: SourceType.
        :return: List of available speeds :rtype: str list.
        """
        if source == SourceType.serial:
            return SerialProcess.get_speeds()
        elif source == SourceType.calibration:
            return CalibrationProcess.get_speeds()
        elif source == SourceType.SocketClient:
            return SocketProcess.get_default_port()
        else:
            print(TAG,'Unknown source selected')
            Log.w(TAG, "Unknown source selected")
            return None
        
    
    ###########################################################################
    # Setup/Clear the internal buffers
    ###########################################################################
    def reset_buffers(self, samples):
        #:param samples: Number of samples for the buffers :type samples: int.
        
        # Initialises data buffers
        self._data1_buffer = np.zeros(samples) # amplitude
        self._data2_buffer = np.zeros(samples) # phase
        #self._d1_buffer = []  # Resonance frequency 
        #self._d2_buffer = []  # Dissipation
        #self._d3_buffer = []  # temperature
        #self._t1_buffer = []  # time (Resonance frequency)
        #self._t2_buffer = []  # time (Dissipation)
        #self._t3_buffer = []  # time (temperature)

        # Initialises supporting variables
        self._d1_store = 0
        self._d2_store = 0
        self._d3_store = 0
        self._t1_store = 0
        self._t2_store = 0
        self._t3_store = 0
        self._ser_error1 = 0
        self._ser_error2 = 0
        self._ser_err_usb= 0
        #self._control_k = 0
        
        self._d1_buffer = RingBuffer(Constants.ring_buffer_samples)  # Resonance frequency 
        self._d2_buffer = RingBuffer(Constants.ring_buffer_samples)  # Dissipation
        self._d3_buffer = RingBuffer(Constants.ring_buffer_samples)  # temperature
        self._t1_buffer = RingBuffer(Constants.ring_buffer_samples)  # time (Resonance frequency)
        self._t2_buffer = RingBuffer(Constants.ring_buffer_samples)  # time (Dissipation)
        self._t3_buffer = RingBuffer(Constants.ring_buffer_samples)  # time (temperature)
        #print(TAG,'Buffers cleared')
        #Log.i(TAG, "Buffers cleared") 

    ############################################################################
    # Gets frequency range
    ############################################################################
    
    def get_frequency_range(self):
        
        """
        :param samples: Number of samples for the buffers :type samples: int.
        :return: overtone :type overtone: float.
        :return: frequency range :type readFREQ: float list.
        """
        return self._readFREQ
    
    
    ############################################################################
    # Gets overtones name, value and frequency step
    ############################################################################
    
    def get_overtone(self):
        
        """
        :param samples: Number of samples for the buffers :type samples: int.
        :return: overtone :type overtone: float.
        :return: frequency range :type readFREQ: float list.
        """
        return self._overtone_name,self._overtone_value, self._fStep
    


from openQCM.core.constants import Constants

TAG = "[Switcher]"

###############################################################################
# Switches from overtone frequency to frequency range for 10MHz QC Sensor
# Returns other parameters that are used for processing
###############################################################################
class Overtone_Switcher_10MHz:
    
    def __init__(self,peak_frequencies = None):
        self.peak_frequencies = peak_frequencies
    
    # from fundamental frequency to the 5th overtone  
    def overtone10MHz_to_freq_range(self, argument):
        method_name = 'overtone_' + str(argument)
        # Get the method from 'self'. Default to a lambda.
        method = getattr(self, method_name, lambda: None)
        # Call the method as we return it
        return method()
 
    def overtone_0(self):
        # fundamental frequency
        name = "fundamental"
        start = self.peak_frequencies[0] - Constants.L10_fundamental
        stop  = self.peak_frequencies[0] + Constants.R10_fundamental
        return name ,self.peak_frequencies[0], start, stop, Constants.SG_window_size10_fundamental, Constants.Spline_factor10_fundamental
 
    def overtone_1(self):
        # 3th Overtone
        name = "3th Overtone"
        start = self.peak_frequencies[1] - Constants.L10_3th_overtone
        stop  = self.peak_frequencies[1] + Constants.R10_3th_overtone
        return name, self.peak_frequencies[1], start, stop, Constants.SG_window_size10_3th_overtone, Constants.Spline_factor10_3th_overtone
    
    def overtone_2(self):
        # 5th Overtone
        name = "5th Overtone"
        start = self.peak_frequencies[2] - Constants.L10_5th_overtone
        stop  = self.peak_frequencies[2] + Constants.R10_5th_overtone
        return name, self.peak_frequencies[2], start, stop, Constants.SG_window_size10_5th_overtone, Constants.Spline_factor10_5th_overtone


###############################################################################
# Switches from overtone frequency to frequency range for 5MHz QC Sensor
###############################################################################
class Overtone_Switcher_5MHz:
    
    def __init__(self,peak_frequencies = None):
        self.peak_frequencies = peak_frequencies
    
    # from fundamental frequency to the 9th overtone    
    def overtone5MHz_to_freq_range(self, argument):
        method_name = 'overtone_' + str(argument)
        # Get the method from 'self'. Default to a lambda.
        method = getattr(self, method_name, lambda: None)
        # Call the method as we return it
        return method()
 
    def overtone_0(self):
        # fundamental frequency
        name = "fundamental"
        start = self.peak_frequencies[0] - Constants.L5_fundamental
        stop  = self.peak_frequencies[0] + Constants.R5_fundamental
        return name, self.peak_frequencies[0], start, stop, Constants.SG_window_size5_fundamental, Constants.Spline_factor5_fundamental
 
    def overtone_1(self):
        # 3th Overtone
        name = "3th Overtone"
        start = self.peak_frequencies[1] - Constants.L5_3th_overtone
        stop  = self.peak_frequencies[1] + Constants.R5_3th_overtone
        return name, self.peak_frequencies[1], start, stop, Constants.SG_window_size5_3th_overtone, Constants.Spline_factor5_3th_overtone
    
    def overtone_2(self):
        # 5th Overtone
        name = "5th Overtone"
        start = self.peak_frequencies[2] - Constants.L5_5th_overtone
        stop  = self.peak_frequencies[2] + Constants.R5_5th_overtone
        return name, self.peak_frequencies[2],start,stop, Constants.SG_window_size5_5th_overtone, Constants.Spline_factor5_5th_overtone

    def overtone_3(self):
        # 7th Overtone
        name = "7th Overtone"
        start = self.peak_frequencies[3] - Constants.L5_7th_overtone
        stop  = self.peak_frequencies[3] + Constants.R5_7th_overtone
        return name, self.peak_frequencies[3],start,stop, Constants.SG_window_size5_7th_overtone, Constants.Spline_factor5_7th_overtone

    def overtone_4(self):
        # 9th Overtone
        name = "9th Overtone"
        start = self.peak_frequencies[4] - Constants.L5_9th_overtone
        stop  = self.peak_frequencies[4] + Constants.R5_9th_overtone
        return name, self.peak_frequencies[4], start, stop, Constants.SG_window_size5_9th_overtone, Constants.Spline_factor5_9th_overtone



import numpy as np

"""
http://code.activestate.com/recipes/68429-ring-buffer/
http://stackoverflow.com/questions/4151320/efficient-circular-buffer
"""
import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning) 

############################################################################
# HistoryBuffer: Fixed-size NumPy array ring buffer
############################################################################

class RingBuffer(object):

    #######################
    def __init__(self, size_max, default_value=np.nan, dtype=float):
        # initialization
        self.size_max = size_max
        self._data = np.empty(size_max, dtype=dtype)
        self._data.fill(default_value)
        self.size = 0

    ########################
    def append(self, value):
        # append new data to ring buffer
        self._data = np.roll(self._data, 1)
        self._data[0] = value
        self.size += 1
        if self.size == self.size_max:
            self.__class__ = RingBufferFull

    ########################
    def get_all(self):
        #return a list of elements from the oldest to the newest
        return self._data

    ########################
    def get_partial(self):
        return self.get_all()[0:self.size]

    ########################
    def __getitem__(self, key):
        #get element
        return self._data[key]

    ########################
    def __repr__(self):
        #return string representation
        s = self._data.__repr__()
        s = s + '\t' + str(self.size)
        s = s + '\t' + self.get_all()[::-1].__repr__()
        s = s + '\t' + self.get_partial()[::-1].__repr__()
        return s

#############################################################################
class RingBufferFull(RingBuffer):
    
    def append(self, value):
        #append an element when buffer is full
        self._data = np.roll(self._data, 1)
        self._data[0] = value



import argparse

from openQCM.common.logger import Logger as Log
from openQCM.common.logger import LoggerLevel
from openQCM.core.constants import Constants


TAG = ""#"Arguments"


###############################################################################
# Creates and parses the arguments to be used by the application
###############################################################################
class Arguments:

    
    ###########################################################################
    # Initializes
    ###########################################################################
    def __init__(self):
        self._parser = None


    ###########################################################################
    # Creates and parses the arguments to be used by the application.
    ###########################################################################
    def create(self):
        parser = argparse.ArgumentParser(
                            description='SOFTWARE\nA real time plotting and logging application')
        parser.add_argument("-i", "--info",
                            dest="log_level_info",
                            action='store_true',
                            help="Enable info messages"
                            )

        parser.add_argument("-d", "--debug",
                            dest="log_level_debug",
                            action='store_true',
                            help="Enable debug messages"
                            )

        parser.add_argument("-v", "--verbose",
                            dest="log_to_console",
                            action='store_true',
                            help="Show log messages in console",
                            default=Constants.log_default_console_log
                            )

        parser.add_argument("-s", "--samples",
                            dest="user_samples",
                            default=Constants.argument_default_samples,
                            help="Specify number of sample to show on plot"
                            )
        self._parser = parser.parse_args()


    ###########################################################################
    # Sets the user specified log level
    ###########################################################################
    def set_user_log_level(self):
        if self._parser is not None:
            self._parse_log_level()
        else:
            Log.w(TAG, "Parser was not created!")
            return None

    ###########################################################################
    # Gets the user specified samples to show in the plot
    ###########################################################################
    def get_user_samples(self):
        """
        :return: Samples specified by user, or default value if not specified.
        :rtype: int.
        """
        return int(self._parser.user_samples)

    ###########################################################################
    # Gets the user specified log to console flag
    ###########################################################################
    def get_user_console_log(self):
        """
        :return: True if log to console is enabled.
        :rtype: bool.
        """
        return self._parser.log_to_console

    ###########################################################################
    # Sets the log level depending on user specification:
    # enable or disable log to console
    ###########################################################################
    def _parse_log_level(self):
        log_to_console = self.get_user_console_log()
        level = LoggerLevel.INFO
        if self._parser.log_level_info:
            level = LoggerLevel.INFO
        elif self._parser.log_level_debug:
            level = LoggerLevel.DEBUG
        Log(level, enable_console=log_to_console)




from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Info(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(100, 500)
        MainWindow.setMinimumSize(QtCore.QSize(0, 0))
        MainWindow.setStyleSheet("")
        MainWindow.setTabShape(QtWidgets.QTabWidget.Rounded)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setObjectName("gridLayout_2")
        
        # Setup Information -------------------------------------------------------------------
        self.info1 = QtWidgets.QLabel()
        self.info1.setStyleSheet('background: #008EC0; padding: 1px;')
        self.info1.setText("<font color=#ffffff > Setup Information&nbsp;</font>")
        self.info1.setFixedWidth(250)
        self.info1.setFixedHeight(15)
        self.gridLayout_2.addWidget(self.info1, 0, 0, 1, 1)
        
        # Device Setup -------------------------------------------------------------------------
        self.info1a = QtWidgets.QLabel()
        self.info1a.setStyleSheet('background: white; padding: 1px; border: 1px solid #cccccc')
        self.info1a.setText("<font color=#0000ff > Device Setup</font>")
        self.info1a.setFixedWidth(250)
        #self.info1a.setFixedHeight(22)
        self.gridLayout_2.addWidget(self.info1a, 1, 0, 1, 1)
        
        # Operation Mode -----------------------------------------------------------------------
        self.info11 = QtWidgets.QLabel()
        self.info11.setStyleSheet('background: white; padding: 1px; border: 1px solid #cccccc')
        self.info11.setText("<font color=#0000ff > Operation Mode </font>")
        self.info11.setFixedWidth(250)
        self.gridLayout_2.addWidget(self.info11, 2, 0, 1, 1)
        
        # Data Information ---------------------------------------------------------------------
        self.info = QtWidgets.QLabel()
        self.info.setStyleSheet('background: #008EC0; padding: 1px;')
        self.info.setText("<font color=#ffffff > Data Information&nbsp;</font>")
        self.info.setFixedWidth(250)
        self.info.setFixedHeight(15)                  
        self.gridLayout_2.addWidget(self.info, 3, 0, 1, 1)
        
        # Selected Frequency -------------------------------------------------------------------
        self.info2 = QtWidgets.QLabel()
        self.info2.setStyleSheet('background: white; padding: 1px; border: 1px solid #cccccc')
        self.info2.setText("<font color=#0000ff > Selected Frequency </font>")
        self.info2.setFixedWidth(250)                  
        self.gridLayout_2.addWidget(self.info2, 4, 0, 1, 1)
        
        # Frequency Value ----------------------------------------------------------------------
        self.info6 = QtWidgets.QLabel()
        self.info6.setStyleSheet('background: white; padding: 1px; border: 1px solid #cccccc')
        self.info6.setText("<font color=#0000ff > Frequency Value </font>")
        self.info6.setFixedWidth(250)                  
        self.gridLayout_2.addWidget(self.info6, 5, 0, 1, 1)
        
        # Start Frequency ----------------------------------------------------------------------
        self.info3 = QtWidgets.QLabel()
        self.info3.setStyleSheet('background: white; padding: 1px; border: 1px solid #cccccc')
        self.info3.setText("<font color=#0000ff > Start Frequency </font>")
        self.info3.setFixedWidth(250)                   
        self.gridLayout_2.addWidget(self.info3, 6, 0, 1, 1)
        
        # Stop Frequency -----------------------------------------------------------------------
        self.info4 = QtWidgets.QLabel()
        self.info4.setStyleSheet('background: white; padding: 1px; border: 1px solid #cccccc')
        self.info4.setText("<font color=#0000ff > Stop Frequency </font>")
        self.info4.setFixedWidth(250)                   
        self.gridLayout_2.addWidget(self.info4, 7, 0, 1, 1)
        
        # Frequency Range----------------------------------------------------------------------
        self.info4a = QtWidgets.QLabel()
        self.info4a.setStyleSheet('background: white; padding: 1px; border: 1px solid #cccccc')
        self.info4a.setText("<font color=#0000ff > Frequency Range </font>")
        self.info4a.setFixedWidth(250)                   
        self.gridLayout_2.addWidget(self.info4a, 8, 0, 1, 1)
        
        # Sample Rate----------------------------------------------------------------------
        self.info5 = QtWidgets.QLabel()
        self.info5.setStyleSheet('background: white; padding: 1px; border: 1px solid #cccccc')
        self.info5.setText("<font color=#0000ff > Sample Rate </font>")
        self.info5.setFixedWidth(250)                   
        self.gridLayout_2.addWidget(self.info5, 9, 0, 1, 1)
        
        # Sample Number----------------------------------------------------------------------
        self.info7 = QtWidgets.QLabel()
        self.info7.setStyleSheet('background: white; padding: 1px; border: 1px solid #cccccc')
        self.info7.setText("<font color=#0000ff > Sample Number </font>")
        self.info7.setFixedWidth(250)                 
        self.gridLayout_2.addWidget(self.info7, 10, 0, 1, 1)
        
        # Reference Settings -------------------------------------------------------------------
        self.inforef = QtWidgets.QLabel()
        self.inforef.setStyleSheet('background: #008EC0; padding: 1px;')
        #self.inforef1.setAlignment(QtCore.Qt.AlignCenter) 
        self.inforef.setText("<font color=#ffffff > Reference Settings </font>")
        self.inforef.setFixedHeight(15)
        self.inforef.setFixedWidth(250)                      
        self.gridLayout_2.addWidget(self.inforef, 11, 0, 1, 1)
        
        # Ref. Frequency -----------------------------------------------------------------------
        self.inforef1 = QtWidgets.QLabel()
        self.inforef1.setStyleSheet('background: white; padding: 1px; border: 1px solid #cccccc')
        #self.inforef1.setAlignment(QtCore.Qt.AlignCenter) 
        self.inforef1.setText("<font color=#0000ff > Ref. Frequency </font>")
        self.inforef1.setFixedWidth(250)
        self.gridLayout_2.addWidget(self.inforef1, 12, 0, 1, 1)
        # Ref. Dissipation -----------------------------------------------------------------------
        
        self.inforef2 = QtWidgets.QLabel()
        self.inforef2.setStyleSheet('background: white; padding: 1px; border: 1px solid #cccccc')
        #self.inforef2.setAlignment(QtCore.Qt.AlignCenter) 
        self.inforef2.setText("<font color=#0000ff > Ref. Dissipation </font>")
        self.inforef2.setFixedWidth(250)                     
        self.gridLayout_2.addWidget(self.inforef2, 13, 0, 1, 1)
        
        # Current Data ---------------------------------------------------------------------------
        self.l8 = QtWidgets.QLabel()
        self.l8.setStyleSheet('background: #008EC0; padding: 1px;')
        self.l8.setText("<font color=#ffffff > Current Data </font>")
        self.l8.setFixedHeight(15)                
        self.l8.setFixedWidth(250)
        self.gridLayout_2.addWidget(self.l8, 14, 0, 1, 1) 
        
        # Resonance Frequency -------------------------------------------------------------------
        self.l7 = QtWidgets.QLabel()
        self.l7.setStyleSheet('background: white; padding: 1px; border: 1px solid #cccccc')
        self.l7.setText("<font color=#0000ff >  Resonance Frequency </font>")
        #self.l7.setAlignment(QtCore.Qt.AlignCenter) 
        self.l7.setFixedWidth(250)
        self.gridLayout_2.addWidget(self.l7, 15, 0, 1, 1) 
        
        # Dissipation ---------------------------------------------------------------------------
        self.l6 = QtWidgets.QLabel()
        self.l6.setStyleSheet('background: white; padding: 1px; border: 1px solid #cccccc')
        self.l6.setText("<font color=#0000ff > Dissipation  </font>")
        self.l6.setFixedWidth(250)                
        self.gridLayout_2.addWidget(self.l6, 16, 0, 1, 1) 
        
        # Temperature ---------------------------------------------------------------------------
        self.l6a = QtWidgets.QLabel()
        self.l6a.setStyleSheet('background: white; padding: 1px; border: 1px solid #cccccc')
        self.l6a.setText("<font color=#0000ff >  Temperature </font>")
        self.l6a.setFixedWidth(250)                
        self.gridLayout_2.addWidget(self.l6a, 17, 0, 1, 1)
        #----------------------------------------------------------------------
        
        self.gridLayout.addLayout(self.gridLayout_2, 3, 1, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowIcon(QtGui.QIcon('favicon.ico')) #.png
        MainWindow.setWindowTitle(_translate("MainWindow", "Information GUI"))

'''
if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    MainWindow = QtWidgets.QMainWindow()
    ui = Ui_Info()
    ui.setupUi(MainWindow)
    MainWindow.show()
    sys.exit(app.exec_())
'''
